/* File: drvPM500.c 		   	*/
/* Version: 2.00		   	*/
/* Date Last Modified: 09/14/99	        */


/* Device Driver Support routines for PM500 motor controller */
/*
 *      Original Author: Mark Rivers
 *
 * Modification Log:
 * -----------------
 * .01  11-19-98        mlr     Initial development, based on drvMM4000.c
 * .02  09-14-99        mlr     Modified to support V4.0 of motor record and
 *                              new model of PM500
 */


#include	<vxWorks.h>
#include	<vme.h>
#include	<types.h>
#include	<ctype.h>
#include	<stdioLib.h>
#include	<sysLib.h>
#include	<tickLib.h>
#include	<string.h>
#include	<iv.h>
#include	<taskLib.h>
#include	<wdLib.h>
#include        <rngLib.h>
#include        <rebootLib.h>

#include	<alarm.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<fast_lock.h>
#include	<recSup.h>
#include	<devSup.h>
#include	<dbScan.h>
#include        <devLib.h>
#include        <errMdef.h>
#include	<special.h>
#include	<module_types.h>

#include	"motor.h"
#include	"drvPM500.h"
#include        "gpibIO.h"
#include        "serialIO.h"

#define STATIC static

#define ALL_CARDS -1

#define GPIB_PORT  0
#define RS232_PORT 1

#define WAIT 0
#define FLUSH 1

#define OUTPUT_TERMINATOR "\r"
#define INPUT_TERMINATOR  '\r'

#define SET_BIT_ON(v,a)  v|=(1<<a)
#define SET_BIT_OFF(v,a) v&=~(1<<a)

#define GPIB_TIMEOUT 2000 /* Command timeout in msec */
#define SERIAL_TIMEOUT 2000 /* Command timeout in msec */

#define BUFF_SIZE 100       /* Maximum length of string to/from PM500 */

struct mess_queue
{
    struct mess_node *head;
    struct mess_node *tail;
};


/*----------------debugging-----------------*/

#ifdef NODEBUG
#define Debug(l,f,v) ;
#else
#define Debug(l,f,v) { if(l<=drvPM500Debug) printf(f,v); }
#endif

volatile int drvPM500Debug = 0;

STATIC volatile int motor_scan_rate = SCAN_RATE;


/*----------------motor state info-----------------*/

struct mot_state
{
    int motor_in_motion;	/* count of motors in motion */
    char ident[BUFF_SIZE];	/* identification string for this card */
    int total_axis;		/* total axis on this card */
    int port_type;		/* GPIB_PORT or RS232_PORT */
    struct serialInfo *serialInfo;  /* For RS-232 */
    int serial_card;            /* Card on which Hideos is running */
    char serial_task[20];       /* Hideos task name for serial port */
    struct gpibInfo *gpibInfo;  /* For GPIB */
    int gpib_link;
    int gpib_address;
    struct mess_info
    {
	struct mess_node *motor_motion;	/* in motion, NULL/node */
	int encoder_present;	/* one YES/NO for each axis */
	double position;	/* one pos for each axis */
	double encoder_position;	/* one pos for each axis */
	double velocity;	/* Raw velocity readback */
        double drive_resolution; /* Drive units per encoder pulse */
        double high_limit;      /* High limit in user units */
        double low_limit;       /* Low limit in user units */
	int no_motion_count;
	unsigned long status;	/* one pos for each axis */
    } motor_info[MAX_AXIS];
};

volatile STATIC struct mot_state **motor_state;
STATIC int PM500_num_cards = 0;
STATIC int PM500_num_channels;

STATIC int total_cards;
STATIC int any_motor_in_motion;
STATIC struct mess_queue mess_queue;	/* in message queue head */
STATIC struct mess_queue free_list;

STATIC FAST_LOCK motor_freelist;
STATIC FAST_LOCK motor_queue;
STATIC SEM_ID motor_sem;


/*----------------functions-----------------*/
STATIC int recv_mess(int card, char *com, int flag);
STATIC int send_mess(int card, char *com);
STATIC int process_messages();
STATIC int query_axis(int card);
STATIC struct mess_node *get_head_node();
STATIC char *my_strtok(char *p, char *l, char **tok_save);
STATIC struct mess_node *motor_malloc();
STATIC void start_status(int card);
STATIC int set_status(int card, int signal);
STATIC int motor_free(struct mess_node * node);
STATIC int motor_send(MOTOR_CALL * u_msg);
STATIC int motor_card_info(int card, MOTOR_CARD_QUERY * cq);
STATIC int motor_axis_info(int card, int signal, MOTOR_AXIS_QUERY * aq);
static long report(int level);
static long init();
STATIC int motor_init();

/*----------------functions-----------------*/

struct PM500_support PM500_access = {
    motor_init,
    motor_send,
    motor_free,
    motor_card_info,
    motor_axis_info,
};

struct
{
    long number;
    DRVSUPFUN report;
    DRVSUPFUN init;
} drvPM500 /* drvPM500 */ =
{
    2,
    report,
    init
};


/*********************************************************
 * Print out driver status report
 *********************************************************/
static long report(int level)
{
  int card;

  if (PM500_num_cards <=0)
    printf("    NO PM500 controllers found\n");
  else
    {
      for (card = 0; card < PM500_num_cards; card++)
	  if (motor_state[card])
            switch (motor_state[card]->port_type)
            {
            case RS232_PORT: 
	        printf("    PM500 controller %d port type = RS-232, id: %s \n", 
		   card, 
		   motor_state[card]->ident);
                break;
            case GPIB_PORT:
	        printf("    PM500 controller %d port type = GPIB, id: %s \n", 
		   card, 
		   motor_state[card]->ident);
                break;
            default:
	        printf("    PM500 controller %d port type = Unknown, id: %s \n", 
		   card, 
		   motor_state[card]->ident);
                break;
            }
    }
  return (0);
}


static long init()
{
   /* 
    * We cannot call motor_init() here, because that function can do GPIB I/O,
    * and hence requires that the drvGPIB have already been initialized.
    * That cannot be guaranteed, so we need to call motor_init from device
    * support
    */
    /* Check for setup */
    if (PM500_num_cards <= 0)
    {
	Debug(1, "init: *PM500 driver disabled* \n PM500Setup() is missing from startup script.\n", 0);
	return (ERROR);
    }

    return ((long) 0);
}



/*****************************************************/
/* send/receive messages from the stepper controller */
/* motor_task()			                     */
/*****************************************************/

STATIC int motor_task(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10)
    int a1, a2, a3, a4, a5, a6, a7, a8, a9, a10;
{
    ULONG tick_count, tick_used, tick_curr;
    STATUS sem_ret;
    int wait_time, i;

    tick_used = tickGet();

    while (1)
    {
	tick_curr = tickGet() - tick_used;
	tick_count = (tick_curr > motor_scan_rate)
	    ? NO_WAIT : (motor_scan_rate - tick_curr);
	wait_time = any_motor_in_motion ? (int) tick_count : WAIT_FOREVER;

        /* Wait until we get a message or until is time to poll */
	sem_ret = semTake(motor_sem, wait_time);

	if (sem_ret == ERROR)
	    tick_used = tickGet();


        /* Gets here in all cases - poll timeout or new message received */
	if (any_motor_in_motion)
	{
	    /* Start data area update on motor cards */
	    start_status(ALL_CARDS);

	    for (i = 0; i < PM500_num_cards; i++)
	    {
		if (motor_state[i] &&                 /* card being used */
		    motor_state[i]->motor_in_motion)  /* a motor on this card moving */
		    query_axis(i);
	    }
	}

	if (sem_ret != ERROR)
	    process_messages();
    }
}



/***********************************************************/
/* send query to get all axis in motions position          */
/* query_axis()                                            */
/* This executes in context of motor_task() so it can wait */
/***********************************************************/
STATIC int query_axis(int card)
{
    MOTOR_RETURN *mess_ret;
    register struct mess_node *motor_motion;
    register struct mess_info *motor_info;
    int i;


    /* get the encoder positions for axis in motion */
    for (i = 0; i < motor_state[card]->total_axis; i++)
    {
	motor_info = &(motor_state[card]->motor_info[i]);
	motor_motion = motor_info->motor_motion;

	/*
	 * Check for commanded motion and actual motion or done before
	 * issuing a callback.
	 */
	if (motor_motion && set_status(card, i))
	{
	    motor_motion->position = motor_info->position;
	    motor_motion->encoder_position = motor_info->encoder_position;
	    motor_motion->velocity = motor_info->velocity;
	    motor_motion->status = motor_info->status;

	    mess_ret = (MOTOR_RETURN *) motor_malloc();
	    mess_ret->callback = motor_motion->callback;
	    mess_ret->precord = motor_motion->precord;
	    mess_ret->position = motor_motion->position;
	    mess_ret->encoder_position = motor_motion->encoder_position;
	    mess_ret->velocity = motor_motion->velocity;
	    mess_ret->status = motor_motion->status;
	    mess_ret->type = motor_motion->type;

	    if (motor_motion->status & RA_OVERTRAVEL ||
		motor_motion->status & RA_DONE ||
		motor_motion->status & RA_PROBLEM)
	    {
	       Debug(3, "motor_task: requested motion DONE\n", 0);

		motor_state[card]->motor_in_motion--;

		motor_free(motor_motion);
		motor_motion = (struct mess_node *) NULL;
		motor_info->motor_motion = (struct mess_node *) NULL;
	    }

	    callbackRequest((CALLBACK *) mess_ret);

	    if (motor_state[card]->motor_in_motion == 0)
	    {
		SET_BIT_OFF(any_motor_in_motion, card);
	    }
	}
    }
    return (0);
}


/*********************************************************
 * Read the status and position of all motors on a card
 * start_status(int card)
 *            if card == -1 then start all cards
 *********************************************************/
STATIC void start_status(int card)
{
 /* The PM500 does not have a way to query the position or status of all motors
  * on a controller with a single command.  Each axis must be queried
  * individually.  Thus, there is nothing to do in this function, everything
  * is done in set_status() below 
  */
}


/**************************************************************
 * Read and parse status and position strings for a card and signal
 * set_status()
 ************************************************************/

STATIC int set_status(int card, int signal)
{
    register struct mess_info *motor_info;
    /* Message parsing variables */
    char axis_name = all_axis_names[signal];
    char status_char;
    char dir_char;
    char command[BUFF_SIZE];
    char response[BUFF_SIZE];
    int rtn_state;
    double motorData;

    motor_info = &(motor_state[card]->motor_info[signal]);

    /* Request the status and position of this motor */
    sprintf(command, "%cR;", axis_name);
    send_mess(card, command);
    recv_mess(card, response, WAIT);
    /* The response string is of the format XD+000754.0.
     * However, there is a "bug" in the PM500.  When a move completes the
     * PM500 sends the string "XD" (i.e. no position) and then the
     * normal response. Must read second string when this happens. 
     */
    if (strlen(response) == 2) recv_mess(card, response, WAIT);
    status_char = response[1];
    dir_char = response[2];
    motorData = atof(&response[2]);

    if (status_char == 'B')
        motor_info->status &= ~RA_DONE;
    else
        motor_info->status |= RA_DONE;

    if (status_char == 'E') 
        motor_info->status |= RA_PROBLEM;
    else
        motor_info->status &= ~RA_PROBLEM;

    motor_info->status &= ~RA_OVERTRAVEL;
    motor_info->status |= RA_DIRECTION;
    if (status_char == 'L') {
        motor_info->status |= RA_OVERTRAVEL;
        if (dir_char == '+')
            motor_info->status |= RA_DIRECTION;
        else
            motor_info->status &= ~RA_DIRECTION;
    }

    /* encoder status */
    motor_info->status &= ~EA_SLIP;
    motor_info->status &= ~EA_POSITION;
    motor_info->status &= ~EA_SLIP_STALL;
    motor_info->status &= ~EA_HOME;

    if (motorData == motor_info->position) 
	motor_info->no_motion_count++;
    else 
	{
	    motor_info->position = motorData;
            motor_info->encoder_position = motorData;
	    motor_info->no_motion_count = 0;
	}

    /* Parse motor velocity? */
    /* NEEDS WORK */

    motor_info->velocity = 0.;

    if (!(motor_info->status & RA_DIRECTION))
	motor_info->velocity *= -1;

    rtn_state = (!motor_info->no_motion_count ||
     (motor_info->status & (RA_OVERTRAVEL | RA_DONE | RA_PROBLEM))) ? 1 : 0;

    return (rtn_state);
}


/*****************************************************/
/* Process messages */
/* process_messages()     */
/*****************************************************/
STATIC int process_messages()
{
    register struct mess_node *motor_motion;
    struct mess_node *node;
    char response[BUFF_SIZE];

    while (node = get_head_node())
    {
	Debug(6, "Got message (%s)\n", node->message);

	/* Check that card and signal exists - check tightened 6/11/96 (jps) */
	if ((node->card >= 0 && node->card < total_cards) &&
	    motor_state[node->card] &&
	    (node->signal >= 0 && node->signal < motor_state[node->card]->total_axis))
	{
	    switch (node->type)
	    {
	    case QUERY:
		/* send a message to PM500 */
		send_mess(node->card, node->message);
		recv_mess(node->card, node->message, WAIT);
		callbackRequest((CALLBACK *) node);
		break;
	    case VELOCITY:
		/* send a message to PM500 */
		motor_motion =
		    motor_state[node->card]->motor_info[node->signal].motor_motion;

		send_mess(node->card, node->message);
		recv_mess(node->card, response, WAIT);

		/*
		 * this is tricky - another motion is here there is a very
		 * large assumption being made here: that the person who sent
		 * the previous motion is the same one that is sending this
		 * one, if he weren't, the guy that sent the original would
		 * never get notified of finish motion.  This makes sense in
		 * record processing since only one record can be assigned to
		 * an axis and sent commands to it. An improvement would be
		 * to check and see if the record pointers were the same, if
		 * they were not, then send a finish message to the previous
		 * registered motion guy.
		 */

		if (!motor_motion)	/* if NULL */
		    motor_state[node->card]->motor_in_motion++;
		else
		    motor_free(motor_motion);

		SET_BIT_ON(any_motor_in_motion, node->card);
		motor_state[node->card]->motor_info[node->signal].motor_motion = node;
		break;
	    case MOTION:
		/* send a message to PM500 */
		motor_motion =
		    motor_state[node->card]->motor_info[node->signal].motor_motion;

		send_mess(node->card, node->message);
		recv_mess(node->card, response, WAIT);

		/* this is tricky - see velocity comment */
		if (!motor_motion)	/* if NULL */
		    motor_state[node->card]->motor_in_motion++;
		else
		    motor_free(motor_motion);

		SET_BIT_ON(any_motor_in_motion, node->card);
		motor_state[node->card]->motor_info[node->signal].no_motion_count = 0;
		motor_state[node->card]->motor_info[node->signal].motor_motion = node;
		break;
	    case INFO:
		start_status(node->card);
		set_status(node->card, node->signal);

		node->position =
		    motor_state[node->card]->motor_info[node->signal].position;
		node->encoder_position =
		    motor_state[node->card]->motor_info[node->signal].encoder_position;
		node->status =
		    motor_state[node->card]->motor_info[node->signal].status;

		callbackRequest((CALLBACK *) node);
		break;

	    case MOVE_TERM:
		send_mess(node->card, node->message);
		recv_mess(node->card, response, WAIT);
		motor_free(node);	/* free message buffer */
		break;
	    default:
		/* send a message to PM500 */
		send_mess(node->card, node->message);
		recv_mess(node->card, response, WAIT);
		motor_free(node);	/* free message buffer */
		break;
	    }
	}
	else
	{
	    Debug(1, "send_mess - invalid card #%d\n", node->card);
	    node->position = 0;
	    node->encoder_position = 0;
	    node->velocity = 0;
	    node->status = RA_PROBLEM;
	    callbackRequest((CALLBACK *) node);
	}
    }
    return (0);
}



/*****************************************************/
/* Get a message off the queue */
/* get_head_node()			     */
/*****************************************************/
STATIC struct mess_node *get_head_node()
{
    struct mess_node *node, *save_node;

    Debug(7, "get_head_node: entry\n", 0);
    /* get a message from queue */
    FASTLOCK(&motor_queue);

    node = mess_queue.head;
    save_node = (struct mess_node *) NULL;
    Debug(7, "get_head_node: node = %p\n", node);

    /* delete node from list */
    if (node)
    {				/* tmm added */
	if (node == mess_queue.head)
	{
	    Debug(7, "get_head_node: setting mess_queue.head=%p\n", node->next);
	    mess_queue.head = node->next;
	}
	if (node == mess_queue.tail)
	{
	    Debug(7, "get_head_node: setting mess_queue.tail=%p\n", save_node);
	    mess_queue.tail = save_node;
	}
	if (save_node)
	{
	    Debug(7, "get_head_node: setting save_node->next=%p\n", node->next);
	    save_node->next = node->next;
	}
    }
    FASTUNLOCK(&motor_queue);

    Debug(7, "get_head_node: returning %p\n", node);
    return (node);
}



/*****************************************************/
/* send a message to the PM500 board		     */
/* send_mess()			                     */
/*****************************************************/
STATIC int send_mess(int card, char *input)
{
    char com[BUFF_SIZE];

    /* We have to make a local copy of the string so we can add a terminator
       because the input string could be a constant */
    strcpy(com, input);
    strcat(com, OUTPUT_TERMINATOR);

    /* Check that card exists */
    if (!motor_state[card])
    {
	Debug(1, "send_mess - invalid card #%d\n", card);
	return (-1);
    }


    if(5<=drvPM500Debug) {
        printf("%.2f : send_mess: sending message to card %d", 
                tickGet()/60., card);
        printf(" message = %s\n", com);
    }
    switch (motor_state[card]->port_type)
    {
    case GPIB_PORT:
        gpibIOSend(motor_state[card]->gpibInfo, com, strlen(com),
                            GPIB_TIMEOUT);
        break;
    case RS232_PORT:
        serialIOSend(motor_state[card]->serialInfo, com, strlen(com),
                            SERIAL_TIMEOUT);
        break;
    }

    return (0);
}



/*****************************************************/
/* receive a message from the PM500 board           */
/* recv_mess()			                     */
/*****************************************************/
STATIC int recv_mess(int card, char *com, int flag)
{
    int timeout;
    int len=0;

    /* Check that card exists */
    if (!motor_state[card])
    {
	Debug(1, "resv_mess - invalid card #%d\n", card);
	return (-1);
    }

    switch (motor_state[card]->port_type)
    {
    case GPIB_PORT:
        if (flag == FLUSH)
            timeout = 0;
        else
            timeout = GPIB_TIMEOUT;
        len = gpibIORecv(motor_state[card]->gpibInfo, com, BUFF_SIZE, 
                            INPUT_TERMINATOR, timeout);
        break;
    case RS232_PORT:
        if (flag == FLUSH)
            timeout = 0;
        else
            timeout = SERIAL_TIMEOUT;
        len = serialIORecv(motor_state[card]->serialInfo, com, BUFF_SIZE,
                            INPUT_TERMINATOR, timeout);
        break;
    }

    if (len == 0) com[0] = '\0'; else com[len-1] = '\0';
    if (len > 0) {
        if(5<=drvPM500Debug) {
            printf("%.2f : recv_mess: card %d", tickGet()/60., card);
            printf(" message = \"%s\"\n", com);
        }
    }
    if ((flag == WAIT) && (len == 0)) {
        Debug(1, "ERROR: recv_mess: card %d", card);
        Debug(1, " no response\n", 0);
    }
    return (len);
}



/*****************************************************/
/* Setup system configuration                        */
/* PM500Setup()                                     */
/*****************************************************/
int PM500Setup(int num_cards,	/* maximum number of controllers in system */
            int num_channels,   /* maximum number of channels on any controller            */
	   int scan_rate)	/* polling rate - 1/60 sec units */
{
    int i;

    if (num_cards < 1 || num_cards > PM500_NUM_CARDS)
	PM500_num_cards = PM500_NUM_CARDS;
    else
	PM500_num_cards = num_cards;

    if (num_channels < 1 || num_channels > PM500_NUM_CHANNELS)
	PM500_num_channels = PM500_NUM_CHANNELS;
    else
	PM500_num_channels = num_channels;

    /* Set motor polling task rate */
    if (scan_rate >= 1 && scan_rate <= sysClkRateGet())
	motor_scan_rate = sysClkRateGet() / scan_rate;
    else
	motor_scan_rate = SCAN_RATE;

   /* 
    * Allocate space for motor_state structures.  Note this must be done
    * before PM500Config is called, so it cannot be done in motor_init()
    * This means that we must allocate space for a card without knowing
    * if it really exists, which is not a serious problem
    */
    (void *) motor_state = malloc(PM500_num_cards * sizeof(struct mot_state *));
    for (i=0; i<PM500_num_cards; i++) motor_state[i] = 
                                        (struct mot_state *) NULL;
    return (0);
}


/*****************************************************/
/* Configure a controller                            */
/* PM500Config()                                    */
/*****************************************************/
int PM500Config(int card,	/* card being configured */
            int port_type,      /* GPIB_PORT or RS232_PORT */
	    int addr1,          /* = link for GPIB or hideos_card for RS-232 */
            int addr2)          /* GPIB address or hideos_task */
{

    if (card < 0 || card >= PM500_num_cards)
        return (ERROR);

    Debug(1, "PM500Config: malloc'ing motor_state\n", 0);
    motor_state[card] = (struct mot_state *) malloc(sizeof(struct mot_state));

    switch (port_type)
    {
    case GPIB_PORT:
        motor_state[card]->port_type = port_type;
        motor_state[card]->gpib_link = addr1;
        motor_state[card]->gpib_address = addr2;
        break;
    case RS232_PORT:
        motor_state[card]->port_type = port_type;
        motor_state[card]->serial_card = addr1;
        strcpy(motor_state[card]->serial_task, (char *) addr2);
        break;
    default:
        return (ERROR);
    }
    return (0);
}



/*****************************************************/
/* initialize all software and hardware		     */
/* motor_init()			     */
/*****************************************************/
STATIC int motor_init()
{
    int i, j;
    char buff[BUFF_SIZE];
    char *firmware;
    int total_axis = 0;
    int status;
    double res=0.;

    /* Check for setup */
    if (PM500_num_cards <= 0)
    {
	Debug(1, "motor_init: *PM500 driver disabled* \n PM500Setup() is missing from startup script.\n", 0);
	return (ERROR);
    }

    for (i = 0; i < PM500_num_cards; i++)
    {
        if (!motor_state[i]) continue;

        total_cards = i+1;
	Debug(2, "motor_init: controller %d\n", i);
        /* Initialize communications channel */
        switch (motor_state[i]->port_type)
        {
        case GPIB_PORT:
            motor_state[i]->gpibInfo = gpibIOInit(
                                motor_state[i]->gpib_link, 
                                motor_state[i]->gpib_address);
            break;
        case RS232_PORT:
            motor_state[i]->serialInfo = serialIOInit(
                                motor_state[i]->serial_card,
                                motor_state[i]->serial_task);
            break;
        }

        /* flush any junk at input port - should not be any data available */
        do recv_mess(i, buff, FLUSH); while (strlen(buff) != 0);

	/* Send a SCUM 1 command to put device in this mode. */
        send_mess(i, "SCUM 1");
        recv_mess(i, buff, WAIT);
        
	/* Set up basic controller parameters 
         *  "ENAINT $B7" means the following:
         *   Bit 0=1, only affected axis halts on limit
         *   Bit 1=1, No message when moving axis beyond limit
         *   Bit 2=1, No query echo, prepends status character to axis.
         *   Bit 3=0, All status returns have status character after axis
         *   Bit 4=0, No acknowledgement when command is received
         *   Bit 5=1, Disable sign-on message at power-up
         *   Bit 6=1, No echo
         *   Bit 7=1, CR terminator only on commands and responses
         *   Bit 8=0, CR terminator only on commands and responses
         *   Bit 9=0, No EOI sent
         *   Bit 10=0, CR terminator only on commands and responses
         *   Bit 11=0, CR terminator only on commands and responses
         *   Bit 12=0, Decimal number format 
         *   Bit 13=0, Eearly serial poll mapping
         *   Bit 14=0, No SRQ assertion
         */
        send_mess(i, "SENAINT $B7");
        recv_mess(i, buff, WAIT);

        /* Send a message and read response from controller to see if 
         * it exists */
        send_mess(i, "SVN?");
        status = recv_mess(i, buff, WAIT);  
        /* Return value is length of response string */
        if (status > 0)
        {
	    motor_state[i]->motor_in_motion = 0;

            /* Don't turn on motor power, too dangerous */
	    /* send_mess(i, MOTOR_ON); */
	    send_mess(i, "SVN?");    /* Read firmware revision string */
	    recv_mess(i, buff, WAIT);
            strcpy(motor_state[i]->ident, &buff[2]);  /* Skip "XD" */
	    Debug(5, "Identification = %s\n", motor_state[i]->ident);

            /* Figure out how many axes this controller has.
             * Do this by querying status of each axis in order */
            for (j=0; j<PM500_num_channels; j++) {
                sprintf(buff, "%cSTAT?", all_axis_names[j]);
                send_mess(i, buff);
                recv_mess(i, buff, WAIT);
                if (buff[1] == 'E') break;
                /* Determine axis type and resolution.
                 * This is a real pain, since the only way to get this
                 * information is to ask for the axis firmware and use a
                 * lookup table.  We only have translators, so I don't
                 * really know how to tell them apart.  Our translators
                 * return 302 (50nm stages) or 309 (25nm stages) at the end of
                 * the configuration string.
                 * I assume rotators return something different, but I don't
                 * know what it is.
                 */
                sprintf(buff, "%cCONFIG?", all_axis_names[j]);
	        send_mess(i, buff);
	        recv_mess(i, buff, WAIT);
                firmware = &buff[8];
	        Debug(3, "motor_init: firmware = %s\n", firmware);
                if (!strcmp(firmware, "302")) {
                    /* 50 nm translator */
                    res = .01;
	            Debug(3, "motor_init: axis %d is a 50 nm translator\n", i);
                } else if (!strcmp(firmware, "309")) {
                    /* 25 nm translator */
                    res = .01;
	            Debug(3, "motor_init: axis %d is a 25 nm translator\n", i);
                } else if (!strcmp(firmware, "XXX")) {
                    /* Rotator */
                    res = .01;
                }
		motor_state[i]->motor_info[j].drive_resolution = res;
                Debug(7, " drive resolution = %f\n", res);
                /* Set the limit fields to 0, can't read from controller */
		motor_state[i]->motor_info[j].low_limit = 0.;
		motor_state[i]->motor_info[j].high_limit = 0.;
		motor_state[i]->motor_info[j].motor_motion = NULL;
            }
    
            total_axis = j;
	    Debug(3, "motor_init: Total axis = %d\n", total_axis);
	    motor_state[i]->total_axis = total_axis;

	    start_status(i);
	    for (j = 0; j < total_axis; j++)
	    {
		motor_state[i]->motor_info[j].status = 0;
		motor_state[i]->motor_info[j].no_motion_count = 0;
		motor_state[i]->motor_info[j].encoder_position = 0;
		motor_state[i]->motor_info[j].position = 0;
		set_status(i, j);  /* Read status of each motor */
	    }

	}
	else
	{
	    Debug(1, "motor_init: Controller NOT found!\n", 0);
	    motor_state[i] = (struct mot_state *) NULL;
	}
    }

    motor_sem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
    any_motor_in_motion = 0;

    FASTLOCKINIT(&motor_queue);
    FASTLOCK(&motor_queue);
    mess_queue.head = (struct mess_node *) NULL;
    mess_queue.tail = (struct mess_node *) NULL;
    FASTUNLOCK(&motor_queue);

    FASTLOCKINIT(&motor_freelist);
    FASTLOCK(&motor_freelist);
    free_list.head = (struct mess_node *) NULL;
    free_list.tail = (struct mess_node *) NULL;
    FASTUNLOCK(&motor_freelist);

    status = taskSpawn("tPM500", 64, VX_FP_TASK | VX_STDIO, 50000, motor_task,
       0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    Debug(1, "Started motor_task, ID=%x, drvPM500 motor_init() complete\n", 
            status);

    return (0);
}



/*****************************************************/
/* send a command to the stepper motor task queue    */
/* motor_send()			                     */
/*****************************************************/
STATIC int motor_send(MOTOR_CALL * u_msg)
{
    struct mess_node *new_message;

    Debug(6, "motor_send invoked, message = (%s)\n", u_msg->message);

    new_message = motor_malloc();
    new_message->callback = u_msg->callback;
    new_message->next = (struct mess_node *) NULL;
    new_message->type = u_msg->type;
    new_message->signal = u_msg->signal;
    new_message->card = u_msg->card;
    new_message->precord = u_msg->precord;
    new_message->status = 0;
    strcpy(new_message->message, u_msg->message);

    FASTLOCK(&motor_queue);

    if (mess_queue.tail)
    {
	mess_queue.tail->next = new_message;
	mess_queue.tail = new_message;
    }
    else
    {
	mess_queue.tail = new_message;
	mess_queue.head = new_message;
    }

    FASTUNLOCK(&motor_queue);

    /* This causes the motor_task() to process the message */
    semGive(motor_sem);
    return (0);
}


/*****************************************************/
/* Tokenize a string                                 */
/*****************************************************/
STATIC char *my_strtok(char *p, char *l, char **tok_save)
{
    char *tok_return;

    if (p)
	*tok_save = p;
    else
	p = *tok_save;

    if (!p)
	return (NULL);

    while (*p && !(index(l, *p)))
	p++;

    if (!(*p))
    {
	tok_return = *tok_save;
	*tok_save = NULL;
	return (tok_return);
    }
    else
	*p = '\0';

    tok_return = *tok_save;
    *tok_save = p + 1;
    return (tok_return);
}


/*****************************************************/
/* malloc a message                                  */
/*****************************************************/
STATIC struct mess_node *motor_malloc()
{
    struct mess_node *node;

    FASTLOCK(&motor_freelist);

    if (!free_list.head)
	node = (struct mess_node *) malloc(sizeof(struct mess_node));
    else
    {
	node = free_list.head;
	free_list.head = node->next;
	if (!free_list.head)
	    free_list.tail = (struct mess_node *) NULL;
    }

    FASTUNLOCK(&motor_freelist);
    return (node);
}


/*****************************************************/
/* free a message                                  */
/*****************************************************/
STATIC int motor_free(struct mess_node * node)
{
    FASTLOCK(&motor_freelist);

    node->next = (struct mess_node *) NULL;

    if (free_list.tail)
    {
	free_list.tail->next = node;
	free_list.tail = node;
    }
    else
    {
	free_list.head = node;
	free_list.tail = node;
    }

    FASTUNLOCK(&motor_freelist);
    return (0);
}



/*---------------------------------------------------------------------*/
/*
 * both of these routines are only to be used at initialization time - before
 * queuing transactions starts for an axis.  These routines do no locking,
 * since transactions will not enter the queue for an axis until after they
 * have run.
 */

/* return the card information to caller */
STATIC int motor_card_info(int card, MOTOR_CARD_QUERY * cq)
{
    if (card >= 0 && 
	card < total_cards && 
	motor_state[card])
    {
	cq->total_axis = motor_state[card]->total_axis;
	cq->card_name = motor_state[card]->ident;
    }
    else
	cq->total_axis = 0;

    return (0);
}

/* return information for an axis to the caller */
STATIC int motor_axis_info(int card, int signal, MOTOR_AXIS_QUERY * aq)
{
    if (card >= 0 && 
	card < total_cards && 
	motor_state[card] &&
	signal >= 0 &&
	signal < motor_state[card]->total_axis)
    {
	aq->position = motor_state[card]->motor_info[signal].position;
	aq->encoder_position =
	    motor_state[card]->motor_info[signal].encoder_position;
	aq->status = motor_state[card]->motor_info[signal].status;
	aq->drive_resolution = 
                motor_state[card]->motor_info[signal].drive_resolution;
	aq->high_limit = 
                motor_state[card]->motor_info[signal].high_limit;
	aq->low_limit = 
                motor_state[card]->motor_info[signal].low_limit;
    }
    else
    {
	aq->position = aq->encoder_position = 0;
	aq->status = RA_PROBLEM;
    }

    return (0);
}

/*---------------------------------------------------------------------*/
